//+------------------------------------------------------------------+
//|                                                        timer.mq4 |
//|                                      Copyright  2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: Multiple timers                                 |
//|      with publisher/subscriber design pattern and abstract class |
//|                        https://www.mql5.com/en/blogs/post/680938 |
//+------------------------------------------------------------------+

#property strict

const int seconds = 2;

// base class for subscribers on timer events
// abstract, can not be instantiated
class TimerNotification
{
  private:
    datetime lastNotification;
    bool checkedNotifyCall;
    
  public:
    TimerNotification(): lastNotification(0), checkedNotifyCall(false)
    {
    }
  
    virtual void notify()
    {
      lastNotification = TimeLocal();
    }
    
    virtual int getSeconds() const = 0; // pure virtual method, must be implemented in descendants
    
    bool isTimeCome()
    {
      if(lastNotification == 0)
      {
        if(!checkedNotifyCall)
        {
          checkedNotifyCall = true;
          return true;
        }
        else
        {
          Print("ERROR: call to TimerNotification::notify() is missing in concrete timer class");
        }
      }
        
      return lastNotification + getSeconds() < TimeLocal();
    }
};

// main and single processor of timer events coming from the terminal, publisher
class MainTimer
{
  private:
    TimerNotification *subscribers[];
  
  public:
    MainTimer()
    {
      EventSetTimer(1);
    }
    
    ~MainTimer()
    {
      EventKillTimer();
    }
    
    void bind(TimerNotification &tn)
    {
      int i, n = ArraySize(subscribers);
      for(i = 0; i < n; ++i)
      {
        if(subscribers[i] == &tn) return;
        if(subscribers[i] == NULL) break;
      }
      
      if(i == n)
      {
        ArrayResize(subscribers, n + 1);
      }
      else
      {
        n = i;
      }
      subscribers[n] = &tn;
    }
    
    void unbind(TimerNotification &tn)
    {
      int n = ArraySize(subscribers);
      for(int i = 0; i < n; ++i)
      {
        if(subscribers[i] == &tn)
        {
          subscribers[i] = NULL;
          return;
        }
      }
    }
    
    void onTimer()
    {
      int n = ArraySize(subscribers);
      for(int i = 0; i < n; ++i)
      {
        if(CheckPointer(subscribers[i]) != POINTER_INVALID)
        {
          if(subscribers[i].isTimeCome())
          {
            subscribers[i].notify();
          }
        }
      }
    }
};

// worker timer class
class Timer: public TimerNotification
{
  private:
    int seconds;
    MainTimer *owner;
    
  public:
    Timer(MainTimer &main, const int s)
    {
      seconds = s;
      owner = &main;
      owner.bind(this);
    }
    
    ~Timer()
    {
      owner.unbind(this);
    }
    
    // we must implement this method from TimerNotification, where it's pure virtual
    virtual int getSeconds() const
    {
      return seconds;
    }
    
    virtual void notify()
    {
      Print(__FUNCSIG__, seconds);
      TimerNotification::notify(); // we must call the base implementation from TimerNotification
    }
};


MainTimer mt();
Timer t1(mt, seconds);
Timer t2(mt, seconds * 2);

void OnTimer()
{
  mt.onTimer();
}

/* OUTPUT EXAMPLE
0	01:40:56.272	timer #ENV6,Daily: void Timer::notify()2
0	01:40:58.301	timer #ENV6,Daily: void Timer::notify()4
0	01:40:59.315	timer #ENV6,Daily: void Timer::notify()2
0	01:41:02.357	timer #ENV6,Daily: void Timer::notify()2
0	01:41:03.414	timer #ENV6,Daily: void Timer::notify()4
0	01:41:05.399	timer #ENV6,Daily: void Timer::notify()2
0	01:41:08.441	timer #ENV6,Daily: void Timer::notify()2
0	01:41:08.441	timer #ENV6,Daily: void Timer::notify()4
*/
